import logging
import gevent
from gevent.queue import Queue
from heapq import heappush, heappop
from datetime import datetime, timedelta
from pyramid.compat import string_types
from pyramid_sockjs.protocol import encode, decode

log = logging.getLogger('pyramid_sockjs')

STATE_NEW = 0
STATE_OPEN = 1
STATE_CLOSING = 2
STATE_CLOSED = 3


class Session(object):
    """ SockJS session object

    ``state``: Session state

    ``manager``: Session manager that hold this session

    ``request``: Request object

    ``registry``: Pyramid component registry

    ``acquierd``: Acquired state, indicates that transport is using session

    ``timeout``: Session timeout

    """

    manager = None
    request = None
    registry = None
    acquired = False
    timeout = timedelta(seconds=10)
    state = STATE_NEW

    def __init__(self, id, timeout=timedelta(seconds=10), request=None):
        self.id = id
        self.expired = False
        self.timeout = timeout
        self.request = request
        self.registry = getattr(request, 'registry', None)
        self.expires = datetime.now() + timeout

        self.queue = Queue()

        self.hits = 0
        self.heartbeats = 0

    def __str__(self):
        result = ['id=%r' % self.id]

        if self.state == STATE_OPEN:
            result.append('connected')
        else:
            result.append('disconnected')

        if self.queue.qsize():
            result.append('queue[%s]' % self.queue.qsize())
        if self.hits:
            result.append('hits=%s' % self.hits)
        if self.heartbeats:
            result.append('heartbeats=%s' % self.heartbeats)

        return ' '.join(result)

    def tick(self, timeout=None):
        self.expired = False

        if timeout is None:
            self.expires = datetime.now() + self.timeout
        else:
            self.expires = datetime.now() + timeout

    def heartbeat(self):
        self.heartbeats += 1

    def expire(self):
        """ Manually expire a session. """
        self.expired = True

    def open(self):
        log.info('open session: %s', self.id)
        self.state = STATE_OPEN
        try:
            self.on_open()
        except:
            log.exception("Exceptin in .on_open method.")

    def close(self):
        """ close session """
        log.info('close session: %s', self.id)
        self.state = STATE_CLOSING
        try:
            self.on_close()
        except:
            log.exception("Exceptin in .on_close method.")

    def closed(self):
        log.info('session closed: %s', self.id)
        self.state = STATE_CLOSED
        self.release()
        self.expire()
        try:
            self.on_closed()
        except:
            log.exception("Exceptin in .on_closed method.")

    def acquire(self, request=None):
        self.manager.acquire(self, request)

    def release(self):
        if self.manager is not None:
            self.manager.release(self)

    def get_transport_message(self, block=True, timeout=None):
        self.tick()
        return self.queue.get(block=block, timeout=timeout)

    def send(self, msg):
        """ send message to client """
        log.info('outgoing message: %s, %s', self.id, msg)
        self.tick()
        self.queue.put_nowait(msg)

    def message(self, msg):
        log.info('incoming message: %s, %s', self.id, msg)
        self.tick()
        try:
            self.on_message(msg)
        except:
            log.exception("Exceptin in .on_message method.")

    def on_open(self):
        """ override in subsclass """

    def on_message(self, msg):
        """ executes when new message is received from client """

    def on_close(self):
        """ executes after session marked as closing """

    def on_closed(self):
        """ executes after session marked as closed """

    def on_remove(self):
        """ executes before removing from session manager """


_marker = object()


class SessionManager(dict):
    """ A basic session manager """

    factory = Session

    _gc_thread = None
    _gc_thread_stop = False

    def __init__(self, name, registry, session=None,
                 gc_cycle=3.0, timeout=timedelta(seconds=10)):
        self.name = name
        self.route_name = 'sockjs-url-%s'%name
        self.registry = registry
        if session is not None:
            self.factory = session
        self.acquired = {}
        self.pool = []
        self.timeout = timeout
        self._gc_cycle = gc_cycle

    def route_url(self, request):
        return request.route_url(self.route_name)

    def start(self):
        if self._gc_thread is None:
            def _gc_sessions():
                while not self._gc_thread_stop:
                    gevent.sleep(self._gc_cycle)
                    self._gc() # pragma: no cover

            self._gc_thread = gevent.Greenlet(_gc_sessions)

        if not self._gc_thread:
            self._gc_thread.start()

    def stop(self):
        if self._gc_thread:
            self._gc_thread_stop = True
            self._gc_thread.join()

    def _gc(self):
        current_time = datetime.now()

        while self.pool:
            expires, session = self.pool[0]

            # check if session is removed
            if session.id in self:
                if expires > current_time:
                    break
            else:
                self.pool.pop(0)
                continue

            expires, session = self.pool.pop(0)

            # Session is to be GC'd immedietely
            if session.expires < current_time:
                if not self.on_session_gc(session):
                    del self[session.id]
                    if session.id in self.acquired:
                        del self.acquired[session.id]
                    if session.state == STATE_OPEN:
                        session.close()
                    if session.state == STATE_CLOSING:
                        session.closed()
                continue

            heappush(self.pool, (session.expires, session))

    def on_session_gc(self, session):
        return session.on_remove()

    def _add(self, session):
        if session.expired:
            raise ValueError("Can't add expired session")

        session.manager = self
        session.registry = self.registry

        self[session.id] = session
        heappush(self.pool, (session.expires, session))

    def get(self, id, create=False, request=None, default=_marker):
        session = super(SessionManager, self).get(id, None)
        if session is None:
            if create:
                session = self.factory(id, self.timeout, request=request)
                self._add(session)
            else:
                if default is not _marker:
                    return default
                raise KeyError(id)

        return session

    def acquire(self, session, request=None):
        sid = session.id

        if sid in self.acquired:
            raise KeyError("Another connection still open")
        if sid not in self:
            raise KeyError("Unknown session")

        session.tick()
        session.hits += 1
        session.manager = self
        if request is not None:
            session.request = request

        self.acquired[sid] = True
        return session

    def is_acquired(self, session):
        return session.id in self.acquired

    def release(self, session):
        if session.id in self.acquired:
            del self.acquired[session.id]

    def active_sessions(self):
        for session in self.values():
            if not session.expired:
                yield session

    def clear(self):
        """ Manually expire all sessions in the pool. """
        while self.pool:
            expr, session = heappop(self.pool)
            if session.state != STATE_CLOSED:
                session.closed()
            del self[session.id]

    def broadcast(self, *args, **kw):
        for session in self.values():
            if not session.expired:
                session.send(*args, **kw)

    def __del__(self):
        self.clear()
        self.stop()



import os
import sys
import logging
import multiprocessing # atexit exception
from setuptools import setup, find_packages

version='0.3.4'

install_requires = ['setuptools',
                    'pyramid >= 1.3',
                    'gevent >= 1.0b2',
                    'gevent-websocket >= 0.3.6',
                    ]

tests_require = install_requires + ['nose']

def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()


setup(name='pyramid_sockjs',
      version=version,
      description=('SockJS server implementation for Pyramid.'),
      long_description='\n\n'.join((read('README.rst'), read('CHANGES.txt'))),
      classifiers=[
          "License :: OSI Approved :: MIT License",
          "Intended Audience :: Developers",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: Implementation :: CPython",
          "Framework :: Pyramid",
          "Topic :: Internet :: WWW/HTTP",
          'Topic :: Internet :: WWW/HTTP :: WSGI'],
      author='Nikolay Kim',
      author_email='fafhrd91@gmail.com',
      url='https://github.com/fafhrd91/pyramid_sockjs/',
      license='MIT',
      packages=find_packages(),
      install_requires = install_requires,
      tests_require = tests_require,
      test_suite = 'nose.collector',
      include_package_data = True,
      zip_safe = False,
      entry_points = {
          'paste.server_runner': [
              'server = pyramid_sockjs.paster:gevent_server_runner',
              ],
          },
      )



import random
import logging
import hashlib
from datetime import datetime, timedelta
from pyramid.response import Response
from pyramid.exceptions import ConfigurationError
from pyramid.httpexceptions import HTTPNotFound, HTTPBadRequest

from pyramid_sockjs.session import SessionManager
from pyramid_sockjs.protocol import json
from pyramid_sockjs.protocol import IFRAME_HTML
from pyramid_sockjs.websocket import init_websocket, init_websocket_hixie
from pyramid_sockjs.transports import handlers
from pyramid_sockjs.transports.utils import session_cookie
from pyramid_sockjs.transports.utils import cors_headers
from pyramid_sockjs.transports.utils import cache_headers
from pyramid_sockjs.transports.websocket import RawWebSocketTransport

log = logging.getLogger('pyramid_sockjs')


def add_sockjs_route(cfg, name='', prefix='/__sockjs__',
                     session=None, session_manager=None,
                     disable_transports=(),
                     sockjs_cdn='http://cdn.sockjs.org/sockjs-0.3.1.min.js',
                     permission=None, decorator=None, cookie_needed=True):
    # set session manager
    if session_manager is None:
        session_manager = SessionManager(name, cfg.registry, session=session)

    if session_manager.name != name:
        raise ConfigurationError(
            "Session manage has to have same name as sockjs route")

    if not hasattr(cfg.registry, '__sockjs_managers__'):
        cfg.registry.__sockjs_managers__ = {}

    if name in cfg.registry.__sockjs_managers__:
        raise ConfigurationError("SockJS '%s' route already registered"%name)

    cfg.registry.__sockjs_managers__[name] = session_manager

    # start gc
    session_manager.start()

    # register routes
    sockjs = SockJSRoute(
        name, session_manager, sockjs_cdn, disable_transports, cookie_needed)

    if prefix.endswith('/'):
        prefix = prefix[:-1]

    route_name = 'sockjs-url-%s-greeting'%name
    cfg.add_route(route_name, prefix)
    cfg.add_view(route_name=route_name, view=sockjs.greeting,
                 permission=permission, decorator=decorator)

    route_name = 'sockjs-url-%s'%name
    cfg.add_route(route_name, '%s/'%prefix)
    cfg.add_view(route_name=route_name, view=sockjs.greeting,
                 permission=permission, decorator=decorator)

    route_name = 'sockjs-%s'%name
    cfg.add_route(route_name, '%s/{server}/{session}/{transport}'%prefix)
    cfg.add_view(route_name=route_name, view=sockjs.handler,
                 permission=permission, decorator=decorator)

    route_name = 'sockjs-websocket-%s'%name
    cfg.add_route(route_name, '%s/websocket'%prefix)
    cfg.add_view(route_name=route_name, view=sockjs.websocket,
                 permission=permission, decorator=decorator)

    route_name = 'sockjs-info-%s'%name
    cfg.add_route(route_name, '%s/info'%prefix)
    cfg.add_view(route_name=route_name, view=sockjs.info,
                 permission=permission, decorator=decorator)

    route_name = 'sockjs-iframe-%s'%name
    cfg.add_route(route_name, '%s/iframe.html'%prefix)
    cfg.add_view(route_name=route_name, view=sockjs.iframe,
                 permission=permission, decorator=decorator)

    route_name = 'sockjs-iframe-ver-%s'%name
    cfg.add_route(route_name, '%s/iframe{version}.html'%prefix)
    cfg.add_view(route_name=route_name, view=sockjs.iframe,
                 permission=permission, decorator=decorator)


class SockJSRoute(object):

    def __init__(self, name, session_manager,
                 sockjs_cdn, disable_transports, cookie_needed=True):
        self.name = name
        self.session_manager = session_manager
        self.disable_transports = dict((k,1) for k in disable_transports)
        self.cookie_needed = cookie_needed
        self.iframe_html = IFRAME_HTML%sockjs_cdn
        self.iframe_html_hxd = hashlib.md5(self.iframe_html).hexdigest()

    def handler(self, request):
        matchdict = request.matchdict

        # lookup transport
        tid = matchdict['transport']

        if tid not in handlers or tid in self.disable_transports:
            return HTTPNotFound()

        create, transport = handlers[tid]

        # session
        manager = self.session_manager

        sid = matchdict['session']
        if not sid or '.' in sid or '.' in matchdict['server']:
            return HTTPNotFound()

        try:
            session = manager.get(sid, create, request=request)
        except KeyError:
            return HTTPNotFound(headers=(session_cookie(request),))

        request.environ['wsgi.sockjs_session'] = session

        # websocket
        if tid == 'websocket':
            if 'HTTP_SEC_WEBSOCKET_VERSION' in request.environ:
                res = init_websocket(request)
            elif 'HTTP_ORIGIN' in request.environ:
                res = init_websocket_hixie(request)
            else:
                res = init_websocket(request)

            if res is not None:
                return res

        try:
            return transport(session, request)
        except Exception as exc:
            session.release()
            log.exception('Exception in transport: %s'%tid)
            return HTTPBadRequest(str(exc))

    def websocket(self, request):
        # session
        manager = self.session_manager

        sid = '%0.9d'%random.randint(1, 2147483647)
        session = manager.get(sid, True, request=request)
        request.environ['wsgi.sockjs_session'] = session

        # websocket
        if 'HTTP_ORIGIN' in request.environ:
            return HTTPNotFound()

        res = init_websocket(request)
        if res is not None:
            return res

        return RawWebSocketTransport(session, request)

    def info(self, request):
        response = request.response
        response.content_type = 'application/json; charset=UTF-8'
        response.headerlist.append(
            ('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0'))
        response.headerlist.extend(cors_headers(request))

        if request.method == 'OPTIONS':
            session_cookie(request)
            response.status = 204
            response.headerlist.append(
                ("Access-Control-Allow-Methods", "OPTIONS, GET"))
            response.headerlist.extend(cache_headers(request))
            return response

        info = {'entropy': random.randint(1, 2147483647),
                'websocket': 'websocket' not in self.disable_transports,
                'cookie_needed': self.cookie_needed,
                'origins': ['*:*']}
        response.body = json.dumps(info)
        return response

    def iframe(self, request):
        response = request.response
        response.headerlist.extend(cache_headers(request))

        cached = request.environ.get('HTTP_IF_NONE_MATCH')
        if cached:
            response.status = 304
            del response.headers['Content-Type']
            return response

        response.headers['Content-Type'] = 'text/html; charset=UTF-8'
        response.headers['ETag'] = self.iframe_html_hxd
        response.body = self.iframe_html
        return response

    def greeting(self, request):
        request.response.content_type = 'text/plain; charset=UTF-8'
        request.response.body = 'Welcome to SockJS!\n'
        return request.response


class GetSessionManager(object):
    """ Pyramid's request.get_sockjs_manager implementation """

    def __init__(self, registry):
        self.registry = registry

    def __call__(self, name=''):
        try:
            return self.registry.__sockjs_managers__[name]
        except AttributeError:
            raise KeyError(name)


def get_session_manager(name, registry):
    try:
        return registry.__sockjs_managers__[name]
    except AttributeError:
        raise KeyError(name)



""" interactive shell cli """
import sys
import code
import json
import Queue
import signal

class ConnectionClosedException(Exception):
    pass


class WebSocket8Client(object):

    def __init__(self, url):
        queue = Queue.Queue()
        self.queue = queue

        from ws4py.client.threadedclient import WebSocketClient
        class IntWebSocketClient(WebSocketClient):
            def received_message(self, m):
                queue.put_nowait(json.loads(str(m)))
            def read_from_connection(self, amount):
                try:
                    r = super(
                        IntWebSocketClient, self).read_from_connection(amount)
                except:
                    queue.put(Ellipsis)
                    return 0
                if not r:
                    queue.put(Ellipsis)
                return r
        self.client = IntWebSocketClient(url)
        self.client.connect()

    def close(self):
        if self.client:
            self.client.running = False
            self.client.close_connection()
            self.client._th.join()
            self.client = None

    def send(self, data):
        self.client.send(data)

    def recv(self):
        try:
            r = self.queue.get(timeout=1.0)
            if r is Ellipsis:
                raise ConnectionClosedException()
            return r
        except Queue.Empty:
            return
        except:
            self.close()
            raise


class InteractiveConsole(code.InteractiveConsole):

    def __init__(self, host):
        code.InteractiveConsole.__init__(self)

        self.host = host
        self.count = 0
        print 'Connecting...'
        self.ws = WebSocket8Client(host)

        r = self.ws.recv()
        if r is not None:
            print r['out'],

    def runsource(self, source, filename='<input>'):
        try:
            if source:
                return self.runsource_ws(source, filename)
        except ConnectionClosedException:
            print "Interactive shell has been disconnected. reconnecting..."
            self.ws = WebSocket8Client(self.host)
            self.ws.recv() # read welcome msg
            return self.runsource_ws(source, filename)

    def runsource_ws(self, source, filename='<input>'):
        self.count += 1

        cmd = json.dumps({'id': self.count, 'source': source})

        self.ws.send(cmd)

        while 1:
            r = self.ws.recv()
            if r is not None:
                if not r['more'] and r['out']:
                    print r['out'],
                if r['complete']:
                    return r['more']
            else:
                pass


def shell():
    console = None
    _handler_int = signal.getsignal(signal.SIGINT)
    _handler_term = signal.getsignal(signal.SIGTERM)

    def process_shutdown(sig, frame):
        if console is not None:
            try:
                console.ws.close()
            except:
                pass

        if sig == signal.SIGINT and callable(_handler_int):
            _handler_int(sig, frame)

        if sig == signal.SIGTERM and callable(_handler_term):
            _handler_term(sig, frame)

        if sig == signal.SIGTERM:
            raise sys.exit()

    signal.signal(signal.SIGINT, process_shutdown)
    signal.signal(signal.SIGTERM, process_shutdown)

    try:
        import readline
    except ImportError:
        pass

    host = sys.argv[1]
    if not host.endswith('/websocket'):
        if host.endswith('/'):
            host = '%swebsocket'%host
        else:
            host = '%s/websocket'%host

    console = InteractiveConsole(host)
    console.interact('')
    console.ws.close()


if __name__ == '__main__':
    shell()
